"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ResolverHandler = void 0;
const ResolvedTimelineHandler_1 = require("./ResolvedTimelineHandler");
const resolvedTimeline_1 = require("../api/resolvedTimeline");
const lib_1 = require("./lib/lib");
const performance_1 = require("./lib/performance");
const timeline_1 = require("./lib/timeline");
const TimelineValidator_1 = require("./TimelineValidator");
/**
 * Note: A Resolver instance is short-lived and used to resolve a timeline.
 * Intended usage:
 * 1. const resolver = new Resolver(options)
 * 2. resolver.run(timeline)
 */
class ResolverHandler {
    constructor(options) {
        this.options = options;
        this.hasRun = false;
        this.nextEvents = [];
        const toc = (0, performance_1.tic)('new Resolver');
        this.resolvedTimeline = new ResolvedTimelineHandler_1.ResolvedTimelineHandler(this.options);
        this.validator = new TimelineValidator_1.TimelineValidator();
        toc();
    }
    /**
     * Resolves a timeline, i.e. resolves the references between objects
     * This method can only be run once per Resolver instance.
     */
    resolveTimeline(timeline) {
        const toc = (0, performance_1.tic)('resolveTimeline');
        /* istanbul ignore if */
        if (this.hasRun)
            throw new Error(`Resolver.resolveTimeline can only run once per instance!
Usage:
const resolver = new Resolver(options);
resolver.run(timeline);`);
        this.hasRun = true;
        // Step 0: Validate the timeline:
        if (!this.options.skipValidation) {
            this.validator.validateTimeline(timeline, false);
        }
        // Step 1: Populate ResolvedTimeline with the timeline:
        for (const obj of timeline) {
            this.resolvedTimeline.addTimelineObject(obj);
        }
        // Step 2: Use cache:
        let cacheHandler;
        if (this.options.cache) {
            cacheHandler = this.resolvedTimeline.initializeCache(this.options.cache);
            cacheHandler.determineChangedObjects();
        }
        // Step 3: Go through and resolve all objects:
        this.resolvedTimeline.resolveAllTimelineObjs();
        // Step 4: Populate nextEvents:
        this.updateNextEvents();
        // Step 5: persist cache
        if (cacheHandler) {
            cacheHandler.persistData();
        }
        const resolvedTimeline = (0, lib_1.literal)({
            objects: (0, lib_1.mapToObject)(this.resolvedTimeline.objectsMap),
            classes: (0, lib_1.mapToObject)(this.resolvedTimeline.classesMap),
            layers: (0, lib_1.mapToObject)(this.resolvedTimeline.layersMap),
            nextEvents: this.nextEvents,
            statistics: this.resolvedTimeline.getStatistics(),
            error: this.resolvedTimeline.resolveError,
        });
        toc();
        return resolvedTimeline;
    }
    /** Update this.nextEvents */
    updateNextEvents() {
        const toc = (0, performance_1.tic)('  updateNextEvents');
        this.nextEvents = [];
        const allObjects = [];
        const allKeyframes = [];
        for (const obj of this.resolvedTimeline.objectsMap.values()) {
            if (obj.resolved.isKeyframe) {
                allKeyframes.push(obj);
            }
            else {
                allObjects.push(obj);
            }
        }
        /** Used to fast-track in cases where there are no keyframes */
        const hasKeyframes = allKeyframes.length > 0;
        const objectInstanceStartTimes = new Set();
        const objectInstanceEndTimes = new Set();
        // Go through keyframes last:
        for (const obj of [...allObjects, ...allKeyframes]) {
            if (!obj.resolved.isKeyframe) {
                if (!(0, timeline_1.objHasLayer)(obj))
                    continue; // transparent objects are omitted in NextEvents
            }
            else if (obj.resolved.parentId !== undefined) {
                const parentObj = this.resolvedTimeline.getObject(obj.resolved.parentId);
                if (parentObj) {
                    /* istanbul ignore if */
                    if (!(0, timeline_1.objHasLayer)(parentObj))
                        continue; // Keyframes of transparent objects are omitted in NextEvents
                }
            }
            for (let i = 0; i < obj.resolved.instances.length; i++) {
                const instance = obj.resolved.instances[i];
                if (instance.start > this.options.time && instance.start < (this.options.limitTime ?? Infinity)) {
                    let useThis = true;
                    if (hasKeyframes) {
                        if (!obj.resolved.isKeyframe) {
                            objectInstanceStartTimes.add(`${obj.id}_${instance.start}`);
                        }
                        else {
                            // No need to put keyframe event if its parent starts at the same time:
                            if (objectInstanceStartTimes.has(`${obj.resolved.parentId}_${instance.start}`)) {
                                useThis = false;
                            }
                        }
                    }
                    if (useThis) {
                        this.nextEvents.push({
                            objId: obj.id,
                            type: obj.resolved.isKeyframe ? resolvedTimeline_1.EventType.KEYFRAME : resolvedTimeline_1.EventType.START,
                            time: instance.start,
                        });
                    }
                }
                if (instance.end !== null &&
                    instance.end > this.options.time &&
                    instance.end < (this.options.limitTime ?? Infinity)) {
                    let useThis = true;
                    if (hasKeyframes) {
                        if (!obj.resolved.isKeyframe) {
                            objectInstanceEndTimes.add(`${obj.id}_${instance.end}`);
                        }
                        else {
                            // No need to put keyframe event if its parent ends at the same time:
                            if (objectInstanceEndTimes.has(`${obj.resolved.parentId}_${instance.end}`)) {
                                useThis = false;
                            }
                        }
                    }
                    if (useThis) {
                        this.nextEvents.push({
                            objId: obj.id,
                            type: obj.resolved.isKeyframe ? resolvedTimeline_1.EventType.KEYFRAME : resolvedTimeline_1.EventType.END,
                            time: instance.end,
                        });
                    }
                }
            }
        }
        this.nextEvents.sort(compareNextEvents);
        toc();
    }
}
exports.ResolverHandler = ResolverHandler;
function compareNextEvents(a, b) {
    return a.time - b.time || b.type - a.type || (0, lib_1.compareStrings)(a.objId, b.objId);
}
//# sourceMappingURL=ResolverHandler.js.map